package File::FindNCD;
require 5.000;
require Exporter;
use Config;
use File::Basename;

=head1 NAME

find - traverse a file tree without changing working directory

finddepth - traverse a directory structure depth-first without
changing working directory

=head1 SYNOPSIS

    use File::FindNCD;
    find \&wanted, '/foo', '/bar';
    sub wanted { ... }
    
    use File::FindNCD;
    finddepth \&wanted, '/foo', '/bar';
    sub wanted { ... }

=head1 DESCRIPTION

The wanted() function performs whatever verifications or functions
you want.  It is passed 3 parameters:

=over 3

=item * the current directory name (e.g. $dir).

=item * the current filename within that directory (e.g. $file).

=item * "$dir/$file" (the fully-qualified filename, e.g. $fullname)

=back

Use the third parameter, $fullname, to access the file when the function
is called (the current directory is not changed).  The function
may return true to prune the tree (only with C<find>, not C<finddepth>).

For example, you can mimic:

    find / -name .nfs\* -mtime +7 \
	-exec rm -f {} \; -o -fstype nfs -prune

with:

    use File::FindNCD;
    find \&wanted, '/foo', '/bar';
    sub wanted {
	my ($dir, $file, $fullname) = @_;
	/^\.nfs.*$/ &&
	(($dev, $ino, $mode, $nlink, $uid, $gid) = lstat $fullname) &&
	int -M _ > 7 &&
	unlink $fullname
	||
	($nlink || (($dev, $ino, $mode, $nlink, $uid, $gid) = lstat $fullname)) &&
	$dev < 0 &&
	(return 1);
    }

Set the variable $File::FindNCD::dont_use_nlink if you're using
AFS, since AFS cheats.

C<finddepth> is just like C<find>, except that it does a depth-first
search.

=head1 NOTE

The purpose of this module, with File::Find already fully functional,
is that there are environments (e.g. where Perl is embedded in a
WWW server such as Microsoft Internet Information Server v2) where
changing the current directory is either impossible or undesirable.

This module is C<use strict> clean.

=head1 AUTHOR

Copyright 1997 Ed Jordan E<lt>ed@fidalgo.netE<gt>, based on File::Find
supplied with Perl 5.

=cut

@ISA = qw(Exporter);
@EXPORT = qw(find finddepth);

use strict; # cunningly and deliberately after the above

my $Is_VMS;

sub find {
    my $wanted = shift;
    my $topdir; # mutter
    foreach $topdir (@_) {
	my ($topdev, $topino, $topmode, $topnlink);
	(($topdev, $topino, $topmode, $topnlink) = stat $topdir)
	  || (warn("Can't stat $topdir: $!\n"), next);
	my ($dir, $file, $fullname);
	if (-d _) {
	    if (-d $topdir) {
		($dir, $file) = ($topdir, '.');
		&$wanted($dir, $file, $topdir);
		$topdir =~ s,/$,,;	# we can fiddle with it as it's
					# no longer needed in its original form
		$topdir =~ s/\.dir$// if $Is_VMS;
		&finddir($wanted, $topdir, $topnlink);
	    } else {
		warn "Can't work with $topdir: $!\n";
	    }
	} else {
	    unless (($dir, $file) = fileparse($topdir)) {
		($dir, $file) = ('.', $topdir);
	    }
	    -d $dir and &$wanted($dir, $file, $topdir);
	}
    }
}

sub finddir {
    my ($wanted, $dir, $nlink) = @_;
    my ($dev, $ino, $mode, $subcount);
    my ($fullname);

    # Get the list of files in the current directory.

    opendir DIR, $dir or (warn "Can't open $dir: $!\n", return);
    my @filenames = readdir DIR;
    closedir DIR;

    my $file;
    if ($nlink == 2 && !$File::FindNCD::dont_use_nlink) {
	# This dir has no subdirectories.
	foreach $file (@filenames) {
	    next if $file eq '.';
	    next if $file eq '..';
	    $fullname = "$dir/$file";
	    &$wanted($dir, $file, $fullname);
	}
    } else {                    # This dir has subdirectories.
	$subcount = $nlink - 2;
	foreach $file (@filenames) {
	    next if $file eq '.';
	    next if $file eq '..';
	    $nlink = 0;
	    $fullname = "$dir/$file";
	    my $prune = &$wanted($dir, $file, $fullname);
	    if ($subcount > 0 || $File::FindNCD::dont_use_nlink) {    # Seen all the subdirs?

		# Get link count and check for directoriness.

		($dev, $ino, $mode, $nlink) = ($Is_VMS ? stat $fullname : lstat $fullname)
		    unless ($nlink || $File::FindNCD::dont_use_nlink);

		if (-d $fullname) { # etj90 - on NT PerlIS, if do "-d _", gives incorrect results
		    # It really is a directory, so do it recursively.

		    if (!$prune && -d $fullname) {
			$fullname =~ s/\.dir$// if $Is_VMS;
			&finddir($wanted, $fullname, $nlink);
		    }
		    --$subcount;
		}
	    }
	}
    }
}

sub finddepth {
    my $wanted = shift;
    my $topdir; # mutter
    foreach $topdir (@_) {
	my ($topdev, $topino, $topmode, $topnlink);
	(($topdev, $topino, $topmode, $topnlink) = stat $topdir)
	  || (warn("Can't stat $topdir: $!\n"), next);
	my ($dir, $file);
	if (-d _) {
	    if (-d $topdir) {
		my $fixtopdir = $topdir;
		$fixtopdir =~ s,/$,, ;
		$fixtopdir =~ s/\.dir$// if $Is_VMS;
		&finddepthdir($wanted, $fixtopdir, $topnlink);
		($dir, $file) = ($fixtopdir, '.');
		&$wanted($dir, $file, $fixtopdir);
	    } else {
		warn "Can't operate on $topdir: $!\n";
	    }
	} else {
	    unless (($dir, $file) = fileparse($topdir)) {
		($dir, $file) = ('.', $topdir);
	    }
	    -d $dir && &$wanted($dir, $file, "$dir/$file");
	}
    }
}

sub finddepthdir {
    my($wanted, $dir, $nlink) = @_;
    my($dev, $ino, $mode, $subcount);
    my($fullname);

    # Get the list of files in the current directory.

    opendir DIR, $dir or warn "Can't open $dir: $!\n";
    my @filenames = readdir DIR;
    closedir DIR;

    my ($file);
    if ($nlink == 2 && !$File::FindNCD::dont_use_nlink) {
	# This dir has no subdirectories.
	foreach $file (@filenames) {
	    next if $file eq '.';
	    next if $file eq '..';
	    &$wanted($dir, $file, "$dir/$file");
	}
    } else {
	# This dir has subdirectories.
	$subcount = $nlink - 2;
	my $prune;
	foreach $file (@filenames) {
	    next if $file eq '.';
	    next if $file eq '..';
	    $fullname = "$dir/$file";
	    if ($subcount > 0 || $File::FindNCD::dont_use_nlink) {    # Seen all the subdirs?

		# Get link count and check for directoriness.

		($dev, $ino, $mode, $nlink) = ($Is_VMS ? stat $fullname : lstat $fullname);

		if (-d $fullname) { # etj90 - on NT PerlIS, if do "-d _", gives incorrect results
		    # It really is a directory, so do it recursively.

		    if (!$prune && -d $fullname) {
			$fullname =~ s/\.dir$// if $Is_VMS;
			&finddepthdir($wanted, $fullname, $nlink);
		    }
		    --$subcount;
		}
	    }
	    $prune = &$wanted($dir, $file, $fullname);
	}
    }
}

if ($Config{'osname'} eq 'VMS') {
  $Is_VMS = 1;
  $File::FindNCD::dont_use_nlink = 1;
}

1;
